# app.py


from . import di

import click
import logging
import sys

HELP_APP = "The title of your app in Bitrise"
HELP_ENDING_DATE = "Ending date to drive the analysis (YYYY-MM-DD)"
HELP_STARTING_DATE = "Starting date to drive the analysis (YYYY-MM-DD)"
HELP_BITRISE_PAT = "A Personal Access Token (PAT) for Bitrise API"
HELP_VELOCITY = "Estimate Bitrise Velocity credits related to builds"
HELP_BUILDS = "Details all statusus for all builds (success, failure and aborted)"
HELP_TIMING = "Expand show also queued and building time for all builds"
HELP_REPORT_STYLE = "The style of report you want (stdout | json | excel). Defaults to stdout"
HELP_TARGET_BRANCH = "The target branch to perform the analysis"

logging.basicConfig(level=logging.INFO)


@click.command()
@click.option("--token", required=True, help=HELP_BITRISE_PAT)
@click.option("--app", required=True, help=HELP_APP)
@click.option("--starting", required=True, help=HELP_STARTING_DATE)
@click.option("--ending", required=True, help=HELP_ENDING_DATE)
@click.option("--detailed-builds", is_flag=True, default=False, help=HELP_BUILDS)
@click.option("--detailed-timing", is_flag=True, default=False, help=HELP_TIMING)
@click.option("--emulate-velocity", is_flag=True, default=False, help=HELP_VELOCITY)
@click.option("--target-branch", required=False, default=None, help=HELP_TARGET_BRANCH)
@click.option("--report-style", required=False, default="stdout", help=HELP_REPORT_STYLE)
def launch(
    token,
    app,
    starting,
    ending,
    detailed_builds,
    detailed_timing,
    emulate_velocity,
    target_branch,
    report_style,
):
    app = di.inject(
        token,
        app,
        starting,
        ending,
        detailed_builds,
        detailed_timing,
        emulate_velocity,
        target_branch,
        report_style,
    )
    app.execute()
    sys.exit(0)



# bitrise.py

from .errors import ErrorCause, BitriseReportsError
from .models import BitriseBuild, BitriseProject, BuildStack, ExecutionStatus
from .models import BuildMachine, BuildMinutes, BitriseWorkflow, MachineSize

from datetime import datetime
import requests
import time

BITRISE_API_URL = "https://api.bitrise.io/v0.1"
FIRST_PAGE = "first-page"
NO_MORE_PAGES = "no-more-pages"


class Bitrise(object):
    def __init__(self, api_token):
        self.api = BitriseApiFetcher(api_token)
        self.converter = RawDataConverter()

    def available_projects(self):
        endpoint = f"{BITRISE_API_URL}/apps"
        raw_data = self.api.get(endpoint)
        return self.converter.projects_from(raw_data)

    def builds_for_project(self, project, starting=None, ending=None):
        endpoint = f"{BITRISE_API_URL}/apps/{project.slug}/builds"
        raw_data = self.api.get(endpoint, starting, ending)
        return self.converter.builds_from(raw_data, project)


class BitriseApiFetcher(object):
    def __init__(self, api_token):
        self.auth = {"Authorization": api_token}

    def get(self, endpoint, starting=None, ending=None):
        results = []
        next = FIRST_PAGE

        params = {}

        if starting:
            params["after"] = self.__unixtime(starting)

        if ending:
            params["before"] = self.__unixtime(ending)

        while next != NO_MORE_PAGES:

            if next != FIRST_PAGE:
                params["next"] = next

            fetched, next_page = self.__get_page(endpoint, params)
            results.extend(fetched)
            next = next_page

        return results

    def __get_page(self, endpoint, args=None):
        response = requests.get(endpoint, headers=self.auth, params=args)

        if response.status_code == requests.codes.ok:
            data = response.json()["data"]
            paging = response.json()["paging"]
            next = paging["next"] if "next" in paging.keys() else NO_MORE_PAGES
            return data, next
        else:
            cause = ErrorCause.NetworkingInfrastructure
            message = f"""
            Error when retriving data from : {endpoint}
            Status = {response.status_code}
            """
            raise BitriseReportsError(cause, message)

    def __unixtime(self, datetime):
        return int(time.mktime(datetime.timetuple()))


class RawDataConverter(object):
    def projects_from(self, json):
        def conversion(json, project):
            return [BitriseProject(item["title"], item["slug"]) for item in json]

        return self.__safely_convert(conversion, json)

    def builds_from(self, json, project):
        def conversion(json, project):
            finished_builds = list(filter(lambda raw: raw["finished_at"] is not None, json))
            return [self.build_from(item, project) for item in finished_builds]

        return self.__safely_convert(conversion, json, project)

    def __safely_convert(self, callable, json, project=None):
        try:
            return callable(json, project)
        except:
            cause = ErrorCause.DataConversion
            message = "Could not parse/convert information from builds"
            raise BitriseReportsError(cause, message)

    def build_from(self, json, project):
        machine = self.machine_from(json["machine_type_id"], json["stack_identifier"])
        workflow = self.workflow_from(json["triggered_workflow"])
        minutes = self.minutes_from(
            json["triggered_at"], json["started_on_worker_at"], json["finished_at"]
        )

        status = self.status_from(json["status"])
        branch = self.branch_from(json["original_build_params"])
        return BitriseBuild(project, machine, workflow, minutes, status, branch)

    def machine_from(self, machine_type_id, stack_identifier):
        size = MachineSize(machine_type_id)
        stack = BuildStack("macos" if "osx" in stack_identifier else "linux")
        return BuildMachine(f"{stack.value}.{size.value}", size, stack)

    def workflow_from(self, triggered_workflow):
        return BitriseWorkflow(triggered_workflow)

    def minutes_from(self, triggered_at, started_at, finished_at):
        triggered = self.__dt(triggered_at)
        started = self.__dt(started_at) if started_at is not None else triggered
        finished = self.__dt(finished_at)
        queued = self.__aproximate_minutes(started - triggered)
        building = self.__aproximate_minutes(finished - started)
        total = queued + building
        return BuildMinutes(queued, building, total)

    def status_from(self, status):
        return ExecutionStatus(status) if status in range(1, 3) else ExecutionStatus.aborted

    def branch_from(self, build_parameters):
        return None if build_parameters is None else build_parameters["branch"]

    def __dt(self, timestamp):
        return datetime.fromisoformat(timestamp.replace("Z", ""))

    def __aproximate_minutes(self, diff):
        return round((diff.days * 24 * 60) + (diff.seconds / 60))



# middleware.py

from .errors import ErrorCause, BitriseReportsError

import logging


class ProjectFinder(object):
    def __init__(self, bitrise):
        self.bitrise = bitrise

    def find(self, name):
        projects = self.bitrise.available_projects()
        target = next(filter(lambda project: project.id == name, projects), None)

        if not target:
            logging.error("Cannot locate an app with name: {name}")
            logging.error("Available projects for this user")
            logging.error(projects)
            cause = ErrorCause.MiddlewareOrchestration
            message = f"{name} not present among projects you have access"
            raise BitriseReportsError(cause, message)

        return target


class BuildsAnalyser(object):
    def __init__(self, bitrise, cruncher, target_branch=None):
        self.bitrise = bitrise
        self.cruncher = cruncher
        self.target_branch = target_branch

    def analyse(self, project, starting=None, ending=None):
        all_builds = self.bitrise.builds_for_project(project, starting, ending)
        target_builds = self.__prune(all_builds)

        if not target_builds:
            message = f"Missing builds for project : {project.id} and branch : {self.target_branch}"
            logging.error(message)
            cause = f"{project.id} has no builds matching criteria"
            message = f"{project.id} has no builds matching criteria"
            raise BitriseReportsError(cause, message)

        breakdowns = [
            self.cruncher.breakdown_per_project(target_builds),
            self.cruncher.breakdown_per_machine(target_builds),
            self.cruncher.breakdown_per_workflow(target_builds),
        ]

        return breakdowns, len(target_builds)

    def __prune(self, builds):
        if self.target_branch is None:
            return builds
        return list(filter(lambda build: build.head_branch == self.target_branch, builds))



# reporting.py

from rich.console import Console
from rich.table import Table
from openpyxl import Workbook

import json
import os


class MetricsReporter:
    def __init__(
        self,
        criteria,
        detailed_builds,
        detailed_timing,
        emulate_velocity,
        report_style,
        console=Console(),
    ):
        self.contextual_delegate = ContextReporter(
            criteria, detailed_builds, detailed_timing, emulate_velocity, console
        )
        self.delegates = {
            "stdout": StdoutReporter(
                criteria, detailed_builds, detailed_timing, emulate_velocity, console
            ),
            "json": JsonReporter(
                criteria, detailed_builds, detailed_timing, emulate_velocity, console
            ),
            "excel": ExcelReporter(
                criteria, detailed_builds, detailed_timing, emulate_velocity, console
            ),
        }
        self.output_delegate = self.__find_delegate(report_style)

    def started(self, app):
        self.contextual_delegate.started(app)

    def feedback(self, project):
        self.contextual_delegate.feedback(project)

    def report(self, breakdowns):
        self.output_delegate.report(breakdowns)

    def __find_delegate(self, strategy):
        return self.delegates[strategy]


class ContextReporter:
    def __init__(self, criteria, detailed_builds, detailed_timing, emulate_velocity, console):
        self.criteria = criteria
        self.detailed_builds = detailed_builds
        self.detailed_timing = detailed_timing
        self.emulate_velocity = emulate_velocity
        self.console = console

    def started(self, app):
        self.__print("\nAnalysing", app, "cyan")

    def feedback(self, project):
        self.__print("\nFound slug", project.slug, "cyan")
        self.__print("Starting", self.__format_date(self.criteria.starting_at), "pink")
        self.__print("Ending  ", self.__format_date(self.criteria.ending_at), "pink")

    def __print(self, prefix, content, color):
        printer = self.console
        printer.print(f"{prefix} → [bold {color}]{content}[/bold {color}]")

    def __format_date(self, datetime):
        return f"{datetime:%A, %d %B (%Y)}"


class StdoutReporter(ContextReporter):
    def report(self, breakdowns):
        printer = self.console
        printer.print("")
        printer.print("Builds processed with success!")
        printer.print(
            "[bold green]NOTE[/bold green] : build times in [bold cyan] minutes [/bold cyan] "
        )

        for breakdown in breakdowns:
            self.report_breakdown(breakdown)

    def report_breakdown(self, breakdown):
        printer = self.console
        printer.print(f"\n[bold green]{breakdown.name}[/bold green]")

        table = Table(show_header=True, header_style="bold magenta")
        table.pad_edge = False
        table.add_column("Name")
        table.add_column("Builds", justify="right")

        if self.detailed_builds:
            table.add_column("Successes", justify="right")
            table.add_column("Failures", justify="right")
            table.add_column("Abortions", justify="right")

        table.add_column("Total time", justify="right")

        if self.detailed_timing:
            table.add_column("Queued time", justify="right")
            table.add_column("Building time", justify="right")

        if self.emulate_velocity:
            table.add_column("Credits Estimation", justify="right")

        sorted_by_total = sorted(
            breakdown.details.items(), key=lambda kv: kv[1].count, reverse=True
        )

        for entry, value in sorted_by_total:
            rows = [entry.id, f"[bold cyan]{value.count}[/bold cyan]"]

            if self.detailed_builds:
                rows.append(f"{value.successes}")
                rows.append(f"{value.failures}")
                rows.append(f"{value.abortions}")

            rows.append(f"[bold cyan]{value.total}[/bold cyan]")
            if self.detailed_timing:
                rows.append(f"{value.queued}")
                rows.append(f"{value.building}")

            if self.emulate_velocity:
                rows.append(f"{value.credits}")

            table.add_row(*rows)

        printer.print(table)


class JsonReporter(ContextReporter):
    def __init__(
        self,
        criteria,
        detailed_builds,
        detailed_timing,
        emulate_velocity,
        console,
        filename="bitrise-metrics.json",
        folder=os.getcwd(),
    ):
        super(JsonReporter, self).__init__(
            criteria, detailed_builds, detailed_timing, emulate_velocity, console
        )
        self.filename = filename
        self.folder = folder

    def report(self, breakdowns):
        data = [self.__process(item) for item in breakdowns]
        path = f"{self.folder}/{self.filename}"

        with open(path, "w") as writer:
            json.dump(data, writer, indent=4)
            self.console.print(f"\nWrote results at [bold green]{path}[/bold green]")

    def __process(self, breakdown):
        flattened = {"description": breakdown.name, "details": []}

        sorted_by_total = sorted(
            breakdown.details.items(), key=lambda kv: kv[1].count, reverse=True
        )

        for analysed, value in sorted_by_total:
            entry = {"name": analysed.id, "count": value.count}

            if self.detailed_builds:
                entry["successes"] = value.successes
                entry["failures"] = value.failures
                entry["abortions"] = value.abortions

            entry["total"] = value.total
            if self.detailed_timing:
                entry["queued"] = value.queued
                entry["building"] = value.building

            if self.emulate_velocity:
                entry["credits"] = value.total.credits

            flattened["details"].append(entry)

        return flattened


class ExcelReporter(ContextReporter):
    def __init__(
        self,
        criteria,
        detailed_builds,
        detailed_timing,
        emulate_velocity,
        console,
        filename="bitrise-metrics.xlsx",
        path=f"{os.getcwd()}/bitrise-metrics.xlsx",
        workbook=Workbook(),
    ):
        super(ExcelReporter, self).__init__(
            criteria, detailed_builds, detailed_timing, emulate_velocity, console
        )
        self.filename = filename
        self.path = path
        self.workbook = workbook
        self.sheet = workbook.active
        self.actual_column_index = 65  # ASCII for "A"

    def report(self, breakdowns):

        self.sheet.column_dimensions["A"].width = 25

        for column in ["A", "B", "C", "D", "E", "F", "G", "H", "I"]:
            self.sheet.column_dimensions[column].auto_size = True
            self.sheet.column_dimensions[column].bestFit = True

        line = 1

        for breakdown in breakdowns:
            self.reset_column_index()
            self.update(line, breakdown.name)
            line = line + 1

            sorted_by_total = sorted(
                breakdown.details.items(), key=lambda kv: kv[1].count, reverse=True
            )

            self.update_and_move_column(line, "Name")

            self.update_and_move_column(line, "Total Builds")
            if self.detailed_builds:
                self.update_and_move_column(line, "Build successes")
                self.update_and_move_column(line, "Build failures")
                self.update_and_move_column(line, "Build abortions")

            self.update_and_move_column(line, "Total time")
            if self.detailed_timing:
                self.update_and_move_column(line, "Queued time")
                self.update_and_move_column(line, "Building time")

            if self.emulate_velocity:
                self.update_and_move_column(line, "Credits estimation")

            for entry, value in sorted_by_total:
                line = line + 1

                self.reset_column_index()
                self.update_and_move_column(line, entry.id)
                self.update_and_move_column(line, value.count)

                if self.detailed_builds:
                    self.update_and_move_column(line, value.successes)
                    self.update_and_move_column(line, value.failures)
                    self.update_and_move_column(line, value.abortions)

                self.update_and_move_column(line, value.total)
                if self.detailed_timing:
                    self.update_and_move_column(line, value.queued)
                    self.update_and_move_column(line, value.building)

                if self.emulate_velocity:
                    self.update_and_move_column(line, value.credits)

            line = line + 2

        self.workbook.save(filename=self.filename)
        self.console.print(f"\nWrote results at [bold green]{self.path}[/bold green]")

    def update(self, line, value):
        self.sheet[f"{chr(self.actual_column_index)}{line}"] = value

    def update_and_move_column(self, line, value):
        self.sheet[f"{chr(self.actual_column_index)}{line}"] = value
        self.actual_column_index = self.actual_column_index + 1

    def reset_column_index(self):
        self.actual_column_index = 65



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bitrise_reports']

package_data = \
{'': ['*']}

install_requires = \
['click==8.0.1',
 'openpyxl==3.0.7',
 'pretty-errors==1.2.21',
 'python-dateutil==2.8.1',
 'requests==2.25.1',
 'rich==10.2.2']

entry_points = \
{'console_scripts': ['bitrise-reports = bitrise_reports:main']}

setup_kwargs = {
    'name': 'bitrise-reports',
    'version': '0.1.0',
    'description': 'The missing tool to extract reports about projects you build on Bitrise',
    'long_description': '# Bitrise Reports\n\n[![Flake8](https://img.shields.io/badge/codestyle-flake8-yellow)](https://flake8.pycqa.org/en/latest/)\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Quality](https://api.codeclimate.com/v1/badges/a9fe25bd995710be45d2/maintainability)](https://codeclimate.com/github/dotanuki-labs/bitrise-reports/maintainability)\n[![Coverage](https://codecov.io/gh/dotanuki-labs/bitrise-reports/branch/main/graph/badge.svg)](https://codecov.io/gh/dotanuki-labs/bitrise-reports)\n[![PyPI](https://img.shields.io/pypi/v/bitrise-reports)](https://pypi.org/project/bitrise-reports/)\n[![Main](https://github.com/dotanuki-labs/bitrise-reports/workflows/Main/badge.svg)](https://github.com/dotanuki-labs/bitrise-reports/actions?query=workflow%3AMain)\n[![License](https://img.shields.io/github/license/dotanuki-labs/bitrise-reports)](https://choosealicense.com/licenses/mit)\n\n## What\n\nA simple cruncher for numbers derived from builds you run on [Bitrise CI](https://www.bitrise.io/). Useful if you are in charge of managing infrastructure capacity related to Bitrise, like detecting/reporting anomalies, evaluating queues impact and so on.\n\nMain features:\n\n- Backed by [Bitrise REST API](https://api-docs.bitrise.io/) under the hood\n- Can compute timing (queued, running and total execution time) for all builds in the given time window\n- Can compute build statuses (success, failure or aborted) for all builds in the given time window\n- Results can be filtered by Git branch (eg **master** or **main**)\n- Result are detailed per machine type and also per Workflow\n- Supports emulation of consumed [Bitrise Velocity credits](https://www.bitrise.io/velocity-plan) (for Enterprise customers)\n- Report types : CLI (stdout), JSON and Excel spreadsheet\n\n## Installing\n\nThis tool requires Python, supporting versions 3.8.x and 3.9.x.\n\nInstall `bitrise-reports` with [pip](https://pypi.org/project/pip/)\n\n```bash\n→ pip install bitrise-reports\n```\n\n## Using\n\nLet\'s say you want analyse numbers for the project `android-flagship`, learning from\nbuilds that ran during April of 2021. You\'ll firstly need a\n[Bitrise Personal Access Token](https://devcenter.bitrise.io/api/authentication/) for\nthat. Note you must be a member in the project you want to analyse.\n\nBy running\n\n```bash\n→ bitrise-reports \\\n    --token=$BITRISE_PAT_TOKEN \\\n    --app=my-app \\\n    --starting=2021-04-01 \\\n    --ending=2021-04-30\n```\nyou should get something like that on your CLI\n\n![](.github/assets/showcase-cli-simple.png)\n\nwhich is a simple overview of what happened.\n\nLet\'s say now that you want to learn about how much time you are spending with queued builds.\n\nYou can run then\n\n```bash\n→ bitrise-reports \\\n    --token=$BITRISE_PAT_TOKEN \\\n    --app=my-app \\\n    --starting=2021-04-01 \\\n    --ending=2021-04-30 \\\n    --detailed-timing\n```\n\nand get a report like this one\n\n![](.github/assets/showcase-cli-timing.png)\n\nLast but not least, suppose you want to learn about execution status for all your Workflows that you run for events in your `master` branch (eg, push or a scheduled build).\n\nYou can run\n\n```bash\n→ bitrise-reports \\\n    --token=$BITRISE_PAT_TOKEN \\\n    --app=my-app \\\n    --starting=2021-04-01 \\\n    --ending=2021-04-30 \\\n    --target-branch=master \\\n    --detailed-builds\n```  \n\nand get a report like about that too\n\n![](.github/assets/showcase-cli-statuses.png)\n\n## Command line interface\n\nThe complete list of CLI options:\n\n| Option           | Details                                                         | Required  |\n|------------------|-----------------------------------------------------------------|-----------|\n| token            | Personal access token for Bitrise API                           | Yes       |\n| app              | The title of your app in Bitrise                                | Yes       |\n| starting         | Starting date in the target time frame                          | Yes       |\n| ending           | Ending date in the target time frame                            | Yes       |\n| detailed-builds  | Details all statuses (success, failure and abortion) for builds | No        |\n| detailed-timing  | Details timing (queued, running, total execution) for builds    | No        |\n| emulate-velocity | Estimate Bitrise Velocity credits consumed                      | No        |\n| target-branch    | Filters build by Git branch                                     | No        |\n| report-style     | The style of report you want                                    | No        |\n\nwhere\n\n- `starting` and `ending` follow **YYYY-MM-DD** convention\n- `report-style` accepts **stdout** (default), **json** or **excel**\n- `detailed-timing` is a CLI flag\n- `detailed-builds` is a CLI flag\n- `emulate-velocity` is a CLI flag\n\nIf you opt-in for a specific report style, the corresponding file - **bitrise-metrics.json** or **bitrise-metrics.xlsx** - will be written in the same folder you are runnint `bitrise-reports`.\n\n## Contributing\n\nIf you want to contribute with this project\n\n- Check the [contribution guidelines](https://github.com/dotanuki-labs/.github/blob/main/CONTRIBUTING.md)\n- Ensure you have Python 3.8.+ installed. I recommend [Pyenv](https://github.com/pyenv/pyenv) for that.\n- Ensure you have [Poetry](https://python-poetry.org/) installed\n- Prepare your environment with [Flake8](https://pypi.org/project/flake8/), [Black](https://pypi.org/project/black/) and [Bandit](https://pypi.org/project/bandit/)\n\n```bash\n→ make setup\n```\n\n- Code you changes\n- Make sure you have a green build\n\n```bash\n→  make inspect test\n```\n\n- Submit your PR 🔥\n\n## Author\n\n- Coded by Ubiratan Soares (follow me on [Twitter](https://twitter.com/ubiratanfsoares))\n\n## License\n\n```\nThe MIT License (MIT)\n\nCopyright (c) 2021 Dotanuki Labs\n\nPermission is hereby granted, free of charge, to any person obtaining a copy of\nthis software and associated documentation files (the "Software"), to deal in\nthe Software without restriction, including without limitation the rights to\nuse, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of\nthe Software, and to permit persons to whom the Software is furnished to do so,\nsubject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS\nFOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR\nCOPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER\nIN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN\nCONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n```\n',
    'author': 'Ubiratan Soares',
    'author_email': 'ubiratanfsoares@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dotanuki-labs/bitrise-reports',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
